/**********************************************************************
 Freeciv - Copyright (C) 1996 - A Kjeldberg, L Gregersen, P Unold
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
***********************************************************************/

/* General functions for handling improvements and their effects - moved
   out of city.c since the effects of improvements can be more than city-wide */

#include <assert.h>
#include <stdlib.h>

#include "city.h"
#include "game.h"
#include "log.h"
#include "map.h"
#include "mem.h"
#include "player.h"
#include "support.h"
#include "unit.h"

#include "improvement.h"

/* get 'struct effect_list' functions: */
#define SPECLIST_TAG effect
#define SPECLIST_TYPE struct effect
#include "speclist_c.h"

/****************************************************************
All the city improvements:
Use get_improvement_type(id) to access the array.
The improvement_types array is now setup in:
   server/ruleset.c (for the server)
   client/packhand.c (for the client)
*****************************************************************/

struct impr_type improvement_types[100];        /* FIXME: Need to make this dynamic */

/* Names of effect ranges.
 * (These must correspond to enum effect_range_id in city.h.)
 */
static char *effect_range_names[] = {
  "None",
  "Building",
  "City",
  "Island",
  "Player",
  "World"
};

/* Names of effect types.
 * (These must correspond to enum effect_type_id in city.h.)
 */
static char *effect_type_names[] = {
  "Adv_Parasite",
  "Airlift",
  "Any_Government",
  "Barb_Attack",
  "Barb_Defend",
  "Capital_City",
  "Capital_Exists",
  "Enable_Nuke",
  "Enable_Space",
  "Enemy_Peaceful",
  "Food_Add_Tile",
  "Food_Bonus",
  "Food_Inc_Tile",
  "Food_Per_Tile",
  "Give_Imm_Adv",
  "Growth_Food",
  "Have_Embassies",
  "Improve_Rep",
  "Luxury_Bonus",
  "Luxury_Pct",
  "Make_Content",
  "Make_Content_Mil",
  "Make_Content_Pct",
  "Make_Happy",
  "May_Declare_War",
  "No_Anarchy",
  "No_Sink_Deep",
  "Nuke_Proof",
  "Pollu_Adj",
  "Pollu_Adj_Pop",
  "Pollu_Adj_Prod",
  "Pollu_Decrease",
  "Prod_Add_Tile",
  "Prod_Bonus",
  "Prod_Inc_Tile",
  "Prod_Per_Tile",
  "Prod_To_Gold",
  "Reduce_Corrupt",
  "Reduce_Waste",
  "Reveal_Cities",
  "Reveal_Map",
  "Revolt_Dist",
  "Science_Bonus",
  "Science_Pct",
  "Size_Unlimit",
  "Slow_Global_Warm",
  "Space_Part",
  "Spy_Resistant",
  "Tax_Bonus",
  "Tax_Pct",
  "Trade_Add_Tile",
  "Trade_Bonus",
  "Trade_Inc_Tile",
  "Trade_Per_Tile",
  "Trade_Route_Pct",
  "Unit_Defend",
  "Unit_Move",
  "Unit_No_Lose_Pop",
  "Unit_Recover",
  "Unit_Repair",
  "Unit_Vet_Combat",
  "Unit_Veteran",
  "Upgrade_Units",
  "Upgrade_One_Step",
  "Upgrade_One_Leap",
  "Upgrade_All_Step",
  "Upgrade_All_Leap",
  "Upkeep_Free"
};

/**************************************************************************
  Convert effect range names to enum; case insensitive;
  returns EFR_LAST if can't match.
**************************************************************************/
Eff_Range_id effect_range_from_str(char *str)
{
  Eff_Range_id ret_id;

  assert(sizeof(effect_range_names)/sizeof(effect_range_names[0])==EFR_LAST);

  for (ret_id = 0; ret_id < EFR_LAST; ret_id++) {
    if (0 == mystrcasecmp(effect_range_names[ret_id], str)) {
      return ret_id;
    }
  }

  return EFR_LAST;
}

/**************************************************************************
  Return effect range name; NULL if bad id.
**************************************************************************/
char *effect_range_name(Eff_Range_id id)
{
  assert(sizeof(effect_range_names)/sizeof(effect_range_names[0])==EFR_LAST);

  if (id < EFR_LAST) {
    return effect_range_names[id];
  } else {
    return NULL;
  }
}

/**************************************************************************
  Convert effect type names to enum; case insensitive;
  returns EFT_LAST if can't match.
**************************************************************************/
Eff_Type_id effect_type_from_str(char *str)
{
  Eff_Type_id ret_id;

  assert(sizeof(effect_type_names)/sizeof(effect_type_names[0])==EFT_LAST);

  for (ret_id = 0; ret_id < EFT_LAST; ret_id++) {
    if (0 == mystrcasecmp(effect_type_names[ret_id], str)) {
      return ret_id;
    }
  }

  return EFT_LAST;
}

/**************************************************************************
  Return effect type name; NULL if bad id.
**************************************************************************/
char *effect_type_name(Eff_Type_id id)
{
  assert(sizeof(effect_type_names)/sizeof(effect_type_names[0])==EFT_LAST);

  if (id < EFT_LAST) {
    return effect_type_names[id];
  } else {
    return NULL;
  }
}

/**************************************************************************
  Returns the index of the improvement which implements the given effect
  type, or B_LAST if none can be found
**************************************************************************/
int find_improvement_with_effect_type(Eff_Type_id type)
{
  int i;
  struct impr_effect *eff;
  for(i=0; i<game.num_improvement_types; i++) {
    for (eff=improvement_types[i].effect;eff->type != EFT_LAST;eff++) {
       if (eff->type == type) return i;
    }
  }
  return B_LAST;
}

int effect_is_activated(struct effect *neweff,
                        struct effect_list *elist[EFR_LAST],
                        struct city *pcity) {
/* Need to check through the elists for effects that this effect is dependent
   on, as well as testing for cond_bldg and cond_gov */
  return 1;
}

void activate_effect(struct effect *neweff,struct effect_list *elist[EFR_LAST],
                     struct city *pcity) {
  neweff->flags |= EF_ACTIVE;
  /* Now we need to check to check for all effects that lie within this
     effect's range, to see if they were dependent on this effect and can
     so now themselves be activated */
}

void add_improvement_effects(struct city *pcity,Impr_Type_id improv) {
  struct effect_list *elist[EFR_LAST];
  struct effect *neweff;
  struct impr_effect *ieff;
  Eff_Range_id range;
  int i;

  for (i=0;i<EFR_LAST;i++) elist[i]=NULL;
  elist[EFR_BUILDING]=elist[EFR_CITY]=&pcity->effects;
  i=map_get_continent(pcity->x,pcity->y);
  elist[EFR_ISLAND]=&(game.continent_effects[i]);
  elist[EFR_PLAYER]=&(city_owner(pcity)->effects);
  elist[EFR_WORLD]=&game.effects;

  for (ieff=improvement_types[improv].effect;
       ieff && ieff->type!=EFT_LAST;ieff++) {
    range=ieff->range;
    if (elist[range]) {
      neweff=fc_calloc(1,sizeof(struct effect));
      neweff->flags=EF_BUILDING;
      neweff->owner=improv;
      neweff->base=ieff;
      effect_list_insert(elist[range],neweff);
      if (effect_is_activated(neweff,elist,pcity)) {
         activate_effect(neweff,elist,pcity);
         freelog(LOG_NORMAL,"Improvement %d: effect added\n",improv);
      }
    } else {
      freelog(LOG_NORMAL,"Improvement %d: effect has invalid range\n",improv);
    }
  }
}
